/* Test: SDIO 4 Bit I/O fr das STM32F103ZET6-Board */

#include "StdTypes.h"
#include "STM32F103.h"
#include "serial.h"
#include "conv.h"
#include "sdio.h"

/* Variablen */
byte  iCardType;     /* siehe "Supported Memory Cards" */
dword CSD_Tab[4];    /* Card Specific Data       */
dword CID_Tab[4];    /* Card Identification Data */
dword RCA;           /* momentane Card Adresse   */
dword SCR;           /* SDCard Conf Register     */

dword LastError;
dword LastResponse;

TSDCARDINFO SDCardInfo;


/* Supported Memory Cards */
#define SD1_1     1
#define SD2_0     2
#define SDHC      3
#define MMC       4
#define SD_IO     5         /* z.Z. nicht detektiert */
#define SDHSMM    6         /* z.Z. nicht detektiert */
#define SDCOMBO   7         /* z.Z. nicht detektiert */
#define SDHC_MMC  8         /* z.Z. nicht detektiert */


/* Konstanten fr dsn SDIO-Anschluss */
/* Flags fr SDIO_SendCommand */
#define WAITRESP_SHORT (1<<6)
#define WAITRESP_LONG  (3<<6)
#define WAITINT        (1<<8)
#define WAITPEND       (1<<9)
#define CPSMEN         (1<<10)
#define SDIOSUSPEND    (1<<11)
#define ENCMDCOMPL     (1<<12)
#define NIEN           (1<<13)
#define CEATACMD       (1<<14)

/*** Bits in SDIO_STA und die ersten 11 auch in SDIO_ICR ***/
#define ALL_STATIC_FLAGS 0x5FF
#define CCRCFAIL           1
#define DCRCFAIL       (1<<1)
#define CTIMEOUT       (1<<2)
#define DTIMEOUT       (1<<3)
#define TXUNDERR       (1<<4)
#define RXOVERR        (1<<5)
#define CMDREND        (1<<6)
#define CMDSENT        (1<<7)
#define DATAEND        (1<<8)
#define STBITERR       (1<<9)
#define DBCKEND        (1<<10)

#define CMDACT         (1<<11)
#define TXACT          (1<<12)
#define RXACT          (1<<13)
#define TXFIFOHE       (1<<14)
#define RXFIFOHF       (1<<15)
#define TXFIFOF        (1<<16)
#define RXFIFOF        (1<<17)
#define TXFIFOE        (1<<18)
#define RXFIFOE        (1<<19)
#define TXDAVL         (1<<20)
#define RXDAVL         (1<<21)
#define SDIOIT         (1<<22)
#define CEATAEND       (1<<23)

/* Zustze fr LastError */
#define WRONG_CMD      (1<<26)



/* Bits in SDIO_RESP1 */
#define SD_OCR_ERRORBITS  0xFFF90088  /* war 0xFDFFE008 */


void SDIO_SendCommand (int Cmd, dword Flags, dword Argument)
{ SDIO_ARG = Argument;
  SDIO_CMD = Cmd | Flags;
}


void _SDIO_DataConfig(dword timeout, dword datalen, dword datactrl)
{ dword tmp;

  SDIO_DTIMER = timeout;   /* Set the SDIO Data TimeOut value */
  SDIO_DLEN   = datalen;   /* Set the SDIO DataLength value */

  tmp = SDIO_DCTRL & 0xFFFFFF08;  /* Get the SDIO DCTRL value */
                                  /* Clear DEN, DTMODE, DTDIR and DBCKSIZE bits */
                                  /* Set DEN bit according to SDIO_DPSM value */
                                  /* Set DTMODE bit according to SDIO_TransferMode value */
                                  /* Set DTDIR bit according to SDIO_TransferDir value */
                                  /* Set DBCKSIZE bits according to SDIO_DataBlockSize value */

  SDIO_DCTRL = tmp | datactrl;    /* Write to SDIO DCTRL */
}


/* Checks for error conditions for R1. */
bool CmdResp1Result(byte cmd)
{ dword erg;
  long i;

  i = 10000;
  erg = 0;
  while ((i) && (!(erg & (CCRCFAIL | CMDREND | CTIMEOUT))))
  { erg = SDIO_STA; --i;  }
  if (!i) erg |= CTIMEOUT;

  LastError = erg;
  LastResponse = 0;

  if (erg & CTIMEOUT)
  { SDIO_ICR = CTIMEOUT;
    return 0;
  }

  if (erg & CCRCFAIL)
  { SDIO_ICR = CCRCFAIL;
    return 0;
  }

  /* Check response received is of desired command */
  erg = SDIO_RESPCMD & 0x3F;
  if (cmd != erg)
  { LastError |= WRONG_CMD;
    return 0;
  }

  /* Clear all the static flags */
  SDIO_ICR = ALL_STATIC_FLAGS;

  /* We have received response, retrieve it for analysis  */
  erg = SDIO_RESP1;
  LastResponse = erg;
  if (erg  & SD_OCR_ERRORBITS)
  { return 0;
  }
  return 1;
}


/* Checks for error conditions for R2 (CID or CSD)  */
bool CmdResp2Result(void)
{ dword erg;
  long i;

  i = 10000;
  erg = 0;
  while ((i) && (!(erg & (CCRCFAIL | CMDREND | CTIMEOUT))))
  { erg = SDIO_STA; --i;  }
  if (!i) erg |= CTIMEOUT;

  if (erg & (CTIMEOUT | CCRCFAIL))
  { SDIO_ICR = erg & (CTIMEOUT | CCRCFAIL);
    LastError = erg;
    LastResponse = 0;
    return 0;
  }

  /* Clear all the static flags */
  SDIO_ICR = ALL_STATIC_FLAGS;
  return 1;
}

/* Masks for R6 Response */
#define SD_R6_GENERAL_UNKNOWN_ERROR     0x00002000
#define SD_R6_ILLEGAL_CMD               0x00004000
#define SD_R6_COM_CRC_FAILED            0x00008000

/* Checks for error conditions for R6 (RCA) */
bool CmdResp6Error(byte cmd, word* prca)
{ dword erg;
  long i;

  i = 10000;
  erg = 0;
  while ((i) && (!(erg & (CCRCFAIL | CMDREND | CTIMEOUT))))
  { erg = SDIO_STA; --i;  }
  if (!i) erg |= CTIMEOUT;

  if (erg & (CTIMEOUT | CCRCFAIL))
  { SDIO_ICR = erg & (CTIMEOUT | CCRCFAIL);
    LastError = erg;
    LastResponse = 0;
    return 0;
  }

  /* Check response received is of desired command */
  if ((SDIO_RESPCMD & 63) != cmd)
  { LastError |= WRONG_CMD; return 0; }

  /* Clear all the static flags */
  SDIO_ICR = ALL_STATIC_FLAGS;

  /* We have received response, retrieve it.  */
  erg = SDIO_RESP1;

  if (erg & (SD_R6_GENERAL_UNKNOWN_ERROR |
             SD_R6_ILLEGAL_CMD |
	     SD_R6_COM_CRC_FAILED))  return 0;

  *prca = (erg >> 16) & 0xFFFF;
  return 1;
}


/************************************************************************************/
/* Initializes the SD Card and put it into StandBy State (Ready for data transfer).
   Wichtig:
   Es gengt nicht, nur die Port-Pins (GPIOx->CRx) richtig aufzusetzen.
   Auf den Pins PC8..PC11 liegen nmlich jeweils 2 unterschiedliche Alternativ-Funktionen.
   PC8+9      --> Timer TIM8 oder SDIO
   PC10..PC11 --> UART4 oder SDIO
   Deshalb MSSEN! TIM8 und UART4 gezielt ausgeschaltet sein, RCC->APBxENR

   Initialisieren der SD-Karte (geht fr SPI-Modus):
   a) MOSI + CS high
   b) mindestens 74 x SCLK pulsen
   c) CS low (= aktiv = select)
   d) Kommando 0 applizieren: GO_IDLE_STATE
   e) Response (Typ=R1, entspricht 1 Byte mit Bit 7 = low) einholen (kann bis zu 8 Bytes dauern)
   f) CS high (= unselect)
   g) Kommando 1 applizieren: SEND_OP_COND
 */

bool SD_Init(void)
{ bool validvoltage;
  dword ACMD41Arg;
  dword erg;
  word irca;
  long i;
                                         /* Portpins und Clocks sind bereits erledigt */
  SDIO_POWER  = SDIO_CLKCR  =
  SDIO_ARG    = SDIO_CMD    =
  SDIO_DTIMER = SDIO_DLEN   =
  SDIO_DCTRL  = 0;

  SDIO_ICR    = 0x00C007FF;
  SDIO_MASK   = 0;

  validvoltage = 0;
  iCardType    = 0;              /* erstmal keine Karte */
  i = 10000; while (--i);

  /* --- Power ON Sequence -------------------------------------------------------*/
  SDIO_POWER = 2;                     /* Powerup, noch ohne Clock zur Karte */
  i = 10000; while (--i);
  SDIO_CLKCR = 178 |(1<<14);          /* HCLK = 72MHz, SDIOCLK = 72MHz, SDIO_CK = HCLK/(178 + 2) = 400 KHz */
  SDIO_POWER = 3;                     /* Set Power State to ON */
  SDIO_CLKCR = (1<<8) |(1<<14) | 178; /* Enable SDIO Clock */
  i = 10000; while (--i);

  /* --- CMD0 = Reset ------------------------------------------------------------*/
  SDIO_SendCommand(0,CPSMEN,0);  /* CMD0: GO_IDLE_STATE, no CMD response required */

  i = 10000; erg = 0;
  while ((i) && (!(erg & CMDSENT)))  { erg = SDIO_STA; --i; }
  if (!i) return 0;              /* TimeOut, kein CMDSENT */
  SDIO_ICR = ALL_STATIC_FLAGS;   /* Clear corresponding bits in SDIO_STA */


  /* --- CMD8: SEND_IF_COND ( gibt es erst ab SD Version 2.0 ) ---------------------*/
  /* Send CMD8 to verify SD card interface operating condition.  */
  /* Argument: - [31:12]: Reserved (shall be set to '0')
               - [11:8]: Supply Voltage (VHS) 0x1 (Range: 2.7-3.6 V)
               - [7:0]: Check Pattern (recommended 0xAA) */
  SDIO_SendCommand(8,WAITRESP_SHORT|CPSMEN,0x1AA);  /* CMD Response--> R7 */

  i = 10000;
  erg = 0;
  while ((i) && (!(erg & (CCRCFAIL | CMDREND | CTIMEOUT)))) { erg = SDIO_STA; --i; }

  if ((!i) || (erg & CTIMEOUT)) /* Card is not V2.0 complient (or wrong voltage)*/
  { SDIO_ICR = CTIMEOUT;        /* Clear corresponding bits in SDIO_STA */
    erg = 0;
  }

  if (erg & CMDREND)         /* Card is SD V2.0 compliant */
  { SDIO_ICR = CMDREND;
    String_Out(" SD compliant 2.0\r\n",2);
    erg = 1;
  }

  ACMD41Arg = 0x80100000;
  if (erg)                      /* es ist eine 2.0 Karte */
  { iCardType = SD2_0;          /* SD Standard Card 2.0  */
    ACMD41Arg |= 0x40000000;
  }
  else                                             /* es ist KEINE 2.0 Karte  */
  { SDIO_SendCommand(55,WAITRESP_SHORT|CPSMEN, 0); /* CMD55 = Vorbereiten "app-specific-cmd" */
    erg = CmdResp1Result(55);                      /* eigentlich 31..16 = RCA, 15..0 leer    */
  }

  SDIO_SendCommand(55,WAITRESP_SHORT|CPSMEN, 0);   /* huch, nochmal CMD55 */
  erg = CmdResp1Result(55);

  /* If erg is 0, it is a MMC card
     If erg is 1, it is a SD card: SD card 2.0 (voltage range mismatch)  or SD card 1.x
   */

  if (!erg)                                          /* MMC CARD */
  { if (LastError != CTIMEOUT)  return(0);
    iCardType = MMC;
  }
  else
  { iCardType = SD1_1;
    i = 64000;                                       /* Send ACMD41 SD_APP_OP_COND with Argument 0x80100000 */
    while ((i) && (!validvoltage))
    { SDIO_SendCommand(55,WAITRESP_SHORT|CPSMEN, 0); /* SEND CMD55 with RCA as 0 */
      erg = CmdResp1Result(55);
      if (!erg)
      {  return(0);
      }
      SDIO_SendCommand(41,WAITRESP_SHORT|CPSMEN,ACMD41Arg);

      i = 10000; erg = 0;
      while ((i) && (!(erg & (CCRCFAIL | CMDREND | CTIMEOUT)))) { erg = SDIO_STA; --i; }
      if ((!i) || (erg & CTIMEOUT))
      { SDIO_ICR = CTIMEOUT;
        return 0;
      }
      SDIO_ICR = ALL_STATIC_FLAGS;

      erg = SDIO_RESP1;
      validvoltage = (bool) (((erg >> 31) == 1) ? 1 : 0);
      --i;
    }
    if (validvoltage)
      String_Out(" voltage OK\r\n",2);
    else
      String_Out(" voltage ERR\r\n",2);

    if (!i) return 0;                               /* invalid Voltage */
    if (erg & 0x40000000) iCardType = SDHC;
  }

  String_Out("CardType = ",2); Long_Out(iCardType,1,2); CRLF_Out(2);

                                                    /* Karte(n) bis zum Standby_State bringen */
  irca = 1;                                         /* hier RCA immer 1 */
  if ((SDIO_POWER & 3) != 3)  return 0;

  if (iCardType != SD_IO)
  { SDIO_SendCommand(2,WAITRESP_LONG|CPSMEN,0);     /* CMD2 fr alle auer SD_IO  */
    if (! CmdResp2Result()) return 0;
    CID_Tab[0] = SDIO_RESP1;
    CID_Tab[1] = SDIO_RESP2;
    CID_Tab[2] = SDIO_RESP3;
    CID_Tab[3] = SDIO_RESP4;

    HexL_Out(CID_Tab[0],2);
    Char_Out('-',2);
    HexL_Out(CID_Tab[1],2);
    Char_Out('-',2);
    HexL_Out(CID_Tab[2],2);
    Char_Out('-',2);
    HexL_Out(CID_Tab[3],2);
    String_Out(" CID\r\n",2);
  }

  if ((SD1_1   == iCardType) ||
      (SD2_0   == iCardType) ||
      (SDCOMBO == iCardType) ||
      (SDHC    == iCardType))
  {                                                /* Send CMD3 SET_REL_ADDR with argument 0 */
    SDIO_SendCommand(3,WAITRESP_SHORT|CPSMEN,0);   /* SD Card publishes its RCA. */
    if (!CmdResp6Error(3, &irca)) return 0;
  }

  if (iCardType != SD_IO)
  { RCA = irca;
                                                   /* Send CMD9 SEND_CSD with argument as card's RCA */
    SDIO_SendCommand(9,WAITRESP_LONG|CPSMEN,(dword)(irca << 16));
    if (!CmdResp2Result()) return 0;
    CSD_Tab[0] = SDIO_RESP1;
    CSD_Tab[1] = SDIO_RESP2;
    CSD_Tab[2] = SDIO_RESP3;
    CSD_Tab[3] = SDIO_RESP4;

    HexL_Out(CSD_Tab[0],2);
    Char_Out('-',2);
    HexL_Out(CSD_Tab[1],2);
    Char_Out('-',2);
    HexL_Out(CSD_Tab[2],2);
    Char_Out('-',2);
    HexL_Out(CSD_Tab[3],2);
    String_Out(" CSD\r\n",2);
  }

  /* HCLK = 72 MHz, SDIOCLK = 72 MHz, SDIO_CK = HCLK/(2 + 1) = 24 MHz */
  /* 1(+2)=24MHz,  BusWide_1b*/
  SDIO_CLKCR = (1<<14) |       /* HW-Flusteuerung ein, um ber/Unterlufe zu vermeiden */
               (1<<8)  |       /* Clock Enable */
               2;              /* Clockrate ist 72MHz/(2+zahl), also hier 72/4=18MHz */

  String_Out("--- SD Init done\r\n",2);
  return 1;    /* All cards get intialized */
}




void Setze_SDIO_CLKCR (byte divisor, byte busbreite)
{ dword tmpreg;

  tmpreg = SDIO_CLKCR;                    /* Get the SDIO CLKCR value */
  tmpreg &= 0xFF00;                       /* Divisor lschen          */
  if (busbreite)
  { tmpreg &= ~(3<<11);                   /* Bits 11+12 lschen       */
    if (busbreite==4) tmpreg |= (1<<11);  /* 01 = 4 Bit, 00 = 1 Bit   */
  }
  tmpreg |= divisor;
  SDIO_CLKCR = tmpreg |(1<<14);           /* Write to SDIO CLKCR */
}



/************************************************************************************/
#define _8Bytes     0x30
#define _sdread     2
#define DPSM_Enable 1

bool FindSCR(word rca)
{ dword erg;
  dword t0;
  dword t1;
  int i;

  SDIO_SendCommand(16,WAITRESP_SHORT|CPSMEN,8); /* Set Block Size to 8 Bytes */
  if (!CmdResp1Result(16)) return 0;
                                                /* Send CMD55 APP_CMD with argument as card's RCA */
  SDIO_SendCommand(55,WAITRESP_SHORT|CPSMEN,(dword) RCA << 16);
  if (!CmdResp1Result(55)) return 0;

  SDIO_DTIMER = 0xFFFF;
  SDIO_DLEN   = 8;
  SDIO_DCTRL  = 3 | (3<<4);   /* blen 8, rd, start */

                                         /* Send ACMD51 SD_APP_SEND_SCR with argument as 0 */
  SDIO_SendCommand(51,WAITRESP_SHORT|CPSMEN,0);  /* war 8);*/
  if (!CmdResp1Result(51)) return 0;

  t0 = t1 = 0;
  /* im FIFO stehen jeweils 4 Bytes - und offenbar in inverser Bytefolge */
  while (!(SDIO_STA & (RXOVERR | DCRCFAIL | DTIMEOUT | DBCKEND | STBITERR)))
  { if (SDIO_STA & RXDAVL)
    { t0 = t1;            /* vom 2. zum 1. */
      t1 = SDIO_FIFO;     /* in den 2.     */
    }
  }
  erg = SDIO_STA;
  LastError = erg;

  erg = erg &(DTIMEOUT | DCRCFAIL | RXOVERR | STBITERR);
  if (erg)
  { SDIO_ICR = erg;
      if (erg & DTIMEOUT) String_Out("Timeout ",2);
      if (erg & DCRCFAIL) String_Out("CrcFail ",2);
      if (erg & RXOVERR)  String_Out("RXover ",2);
      if (erg & STBITERR) String_Out("StBitEr",2);
      CRLF_Out(2);
    return 0;
  }

  /* Clear all the static flags */
  SDIO_ICR = ALL_STATIC_FLAGS;

  /* jeweils:
     0..7   --> 24..31,
     8..15  --> 16..23,
     16..23 --> 8..15,
     24..31 --> 0..7
   */
  i = 4;
  while (i)
  { erg = (erg << 8) | (t0 & 0xFF);
    t0 = t0>>8;
    --i;
  }
  SCR = erg;
  HexL_Out(erg,2); String_Out(" SCR\r\n",2);

  return 1;
}

#define SD_CARD_LOCKED         0x02000000
#define SD_WIDE_BUS_SUPPORT    0x00040000
#define SD_SINGLE_BUS_SUPPORT  0x00010000

bool SwitchToWideBus (void)
{ dword erg;

  erg = SDIO_RESP1;
  LastResponse = erg;
  if (erg & SD_CARD_LOCKED)  return 0;
  if (!FindSCR(RCA))    return 0;   /* Get SCR Register */

  String_Out("SCR ok\r\n",2);
  if ((SCR & SD_WIDE_BUS_SUPPORT)) /* Test, ob Karte wide bus kann */
    {                              /* Send CMD55 APP_CMD with argument as card's RCA.*/
      SDIO_SendCommand(55,WAITRESP_SHORT|CPSMEN,(dword) RCA << 16);
      if (!CmdResp1Result(55)) return 0;
                                      /* Send ACMD6 APP_CMD with argument as 2 for wide bus mode */
      SDIO_SendCommand(6,WAITRESP_SHORT|CPSMEN,2);
      if (!CmdResp1Result(6)) { return 0; }
      return 1;
    }
  return 0;
}


bool SwitchTo1BitBus (void)
{ dword erg;

  erg = SDIO_RESP1;
  LastError = erg;
  if (erg & SD_CARD_LOCKED)  return 0;

  if (!FindSCR(RCA)) return 0;   /* Get SCR Register */
    if ((SCR & SD_SINGLE_BUS_SUPPORT))
    {                        /* Send CMD55 APP_CMD with argument as card's RCA.*/
      SDIO_SendCommand(55,WAITRESP_SHORT|CPSMEN,(dword) RCA << 16);
      if (!CmdResp1Result(55)) return 0;
                            /* Send ACMD6 APP_CMD with argument as 0 for narrow bus mode */
      SDIO_SendCommand(6,WAITRESP_SHORT|CPSMEN,0);
      if (!CmdResp1Result(6)) return 0;
      return 1;
    }
  return 0;
}


/* Selects od Deselects the corresponding card. addr: Address of the Card to be selected. */
bool SD_SelectDeselect(dword addr)
{ SDIO_SendCommand(7,WAITRESP_SHORT|CPSMEN,addr);                 /* Send CMD7 SDIO_SEL_DESEL_CARD */
  return CmdResp1Result(7);
}


/**
  * Checks if the SD card is in programming state.
  *  pstatus: pointer to the variable that will contain the SD card state.
  */
bool IsCardProgramming (byte *pstatus)
{ dword erg;
  long i;

  SDIO_SendCommand(13,WAITRESP_SHORT|CPSMEN,(dword) RCA << 16);
  i = 10000;
  erg = 0;
  while ((i) && (!(erg & (CCRCFAIL | CMDREND | CTIMEOUT))))
  { erg = SDIO_STA; --i;  }

  LastError = erg;
  *pstatus = 0x0F;             /* illegal card state */
  erg = erg & (CCRCFAIL | CTIMEOUT);
  if (erg)
  { SDIO_ICR = erg;
    return 0;
  }
  if ((SDIO_RESPCMD & 0x3F) != 13) return 0;
  SDIO_ICR = ALL_STATIC_FLAGS;

  erg = SDIO_RESP1;
  *pstatus = (byte) ((erg >> 9) & 0x0F);

  if (erg & SD_OCR_ERRORBITS) return 0;
  return 1;
}



/***************************************************************************************/

void TranslateCardInfo(TSDCARDINFO* cardinfo)
{ byte tmp;

  cardinfo->CardType = (byte)iCardType;
  cardinfo->RCA = (word)RCA;

  /* Byte 0 */
  tmp = (byte)((CSD_Tab[0] & 0xFF000000) >> 24);
  cardinfo->SD_csd.CSDStruct = (tmp & 0xC0) >> 6;
  cardinfo->SD_csd.SysSpecVersion = (tmp & 0x3C) >> 2;
  cardinfo->SD_csd.Reserved1 = tmp & 0x03;

  /* Byte 1 */
  tmp = (byte)((CSD_Tab[0] & 0x00FF0000) >> 16);
  cardinfo->SD_csd.TAAC = tmp;

  /* Byte 2 */
  tmp = (byte)((CSD_Tab[0] & 0x0000FF00) >> 8);
  cardinfo->SD_csd.NSAC = tmp;

  /* Byte 3 */
  tmp = (byte)(CSD_Tab[0] & 0x000000FF);
  cardinfo->SD_csd.MaxBusClkFrec = tmp;

  /* Byte 4 */
  tmp = (byte)((CSD_Tab[1] & 0xFF000000) >> 24);
  cardinfo->SD_csd.CardComdClasses = tmp << 4;

  /* Byte 5 */
  tmp = (byte)((CSD_Tab[1] & 0x00FF0000) >> 16);
  cardinfo->SD_csd.CardComdClasses |= (tmp & 0xF0) >> 4;
  cardinfo->SD_csd.RdBlockLen = tmp & 0x0F;

  /* Byte 6 */
  tmp = (byte)((CSD_Tab[1] & 0x0000FF00) >> 8);
  cardinfo->SD_csd.PartBlockRead = (tmp & 0x80) >> 7;
  cardinfo->SD_csd.WrBlockMisalign = (tmp & 0x40) >> 6;
  cardinfo->SD_csd.RdBlockMisalign = (tmp & 0x20) >> 5;
  cardinfo->SD_csd.DSRImpl = (tmp & 0x10) >> 4;
  cardinfo->SD_csd.Reserved2 = 0; /* Reserved */

  if ((iCardType == SD1_1) || (iCardType == SD2_0))
  { cardinfo->SD_csd.DeviceSize = (tmp & 3) << 10;

    /* Byte 7 */
    tmp = (byte)(CSD_Tab[1] & 0x000000FF);
    cardinfo->SD_csd.DeviceSize |= (tmp) << 2;

    /* Byte 8 */
    tmp = (byte)((CSD_Tab[2] & 0xFF000000) >> 24);
    cardinfo->SD_csd.DeviceSize |= (tmp & 0xC0) >> 6;

    cardinfo->SD_csd.MaxRdCurrentVDDMin = (tmp & 0x38) >> 3;
    cardinfo->SD_csd.MaxRdCurrentVDDMax = (tmp & 0x07);

    /* Byte 9 */
    tmp = (byte)((CSD_Tab[2] & 0x00FF0000) >> 16);
    cardinfo->SD_csd.MaxWrCurrentVDDMin = (tmp & 0xE0) >> 5;
    cardinfo->SD_csd.MaxWrCurrentVDDMax = (tmp & 0x1C) >> 2;
    cardinfo->SD_csd.DeviceSizeMul = (tmp & 0x03) << 1;
    /* Byte 10 */
    tmp = (byte)((CSD_Tab[2] & 0x0000FF00) >> 8);
    cardinfo->SD_csd.DeviceSizeMul |= (tmp & 0x80) >> 7;

    cardinfo->CardCapacity = (cardinfo->SD_csd.DeviceSize + 1) ;
    cardinfo->CardCapacity *= (1 << (cardinfo->SD_csd.DeviceSizeMul + 2));
    cardinfo->CardBlockSize = 1 << (cardinfo->SD_csd.RdBlockLen);
    cardinfo->CardCapacity *= cardinfo->CardBlockSize;
  }
  else if (iCardType == SDHC)
  { /* Byte 7 */
    tmp = (byte)(CSD_Tab[1] & 0xFF);
    cardinfo->SD_csd.DeviceSize = (tmp & 0x3F) << 16;

    /* Byte 8 */
    tmp = (byte)((CSD_Tab[2] & 0xFF000000) >> 24);

    cardinfo->SD_csd.DeviceSize |= (tmp << 8);

    /* Byte 9 */
    tmp = (byte)((CSD_Tab[2] & 0x00FF0000) >> 16);

    cardinfo->SD_csd.DeviceSize |= (tmp);

    /* Byte 10 */
    tmp = (byte)((CSD_Tab[2] & 0x0000FF00) >> 8);

    cardinfo->CardCapacity = (cardinfo->SD_csd.DeviceSize + 1) * 512 * 1024;
    cardinfo->CardBlockSize = 512;
  }


  cardinfo->SD_csd.EraseGrSize = (tmp & 0x40) >> 6;
  cardinfo->SD_csd.EraseGrMul = (tmp & 0x3F) << 1;

  /* Byte 11 */
  tmp = (byte)(CSD_Tab[2] & 0x000000FF);
  cardinfo->SD_csd.EraseGrMul |= (tmp & 0x80) >> 7;
  cardinfo->SD_csd.WrProtectGrSize = (tmp & 0x7F);

  /* Byte 12 */
  tmp = (byte)((CSD_Tab[3] & 0xFF000000) >> 24);
  cardinfo->SD_csd.WrProtectGrEnable = (tmp & 0x80) >> 7;
  cardinfo->SD_csd.ManDeflECC = (tmp & 0x60) >> 5;
  cardinfo->SD_csd.WrSpeedFact = (tmp & 0x1C) >> 2;
  cardinfo->SD_csd.MaxWrBlockLen = (tmp & 0x03) << 2;

  /* Byte 13 */
  tmp = (byte)((CSD_Tab[3] & 0x00FF0000) >> 16);
  cardinfo->SD_csd.MaxWrBlockLen |= (tmp & 0xC0) >> 6;
  cardinfo->SD_csd.WriteBlockPaPartial = (tmp & 0x20) >> 5;
  cardinfo->SD_csd.Reserved3 = 0;
  cardinfo->SD_csd.ContentProtectAppli = (tmp & 0x01);

  /* Byte 14 */
  tmp = (byte)((CSD_Tab[3] & 0x0000FF00) >> 8);
  cardinfo->SD_csd.FileFormatGrouop = (tmp & 0x80) >> 7;
  cardinfo->SD_csd.CopyFlag = (tmp & 0x40) >> 6;
  cardinfo->SD_csd.PermWrProtect = (tmp & 0x20) >> 5;
  cardinfo->SD_csd.TempWrProtect = (tmp & 0x10) >> 4;
  cardinfo->SD_csd.FileFormat = (tmp & 0x0C) >> 2;
  cardinfo->SD_csd.ECC = (tmp & 0x03);

  /* Byte 15 */
  tmp = (byte)(CSD_Tab[3] & 0x000000FF);
  cardinfo->SD_csd.CSD_CRC = (tmp & 0xFE) >> 1;
  cardinfo->SD_csd.Reserved4 = 1;


  /* Byte 0 */
  tmp = (byte)((CID_Tab[0] & 0xFF000000) >> 24);
  cardinfo->SD_cid.ManufacturerID = tmp;

  /* Byte 1 */
  tmp = (byte)((CID_Tab[0] & 0x00FF0000) >> 16);
  cardinfo->SD_cid.OEM_AppliID = tmp << 8;

  /* Byte 2 */
  tmp = (byte)((CID_Tab[0] & 0x000000FF00) >> 8);
  cardinfo->SD_cid.OEM_AppliID |= tmp;

  /* Byte 3 */
  tmp = (byte)(CID_Tab[0] & 0x000000FF);
  cardinfo->SD_cid.ProdName[0] = tmp;

  /* Byte 4 */
  tmp = (byte)((CID_Tab[1] & 0xFF000000) >> 24);
  cardinfo->SD_cid.ProdName[1] = tmp;

  /* Byte 5 */
  tmp = (byte)((CID_Tab[1] & 0x00FF0000) >> 16);
  cardinfo->SD_cid.ProdName[2] = tmp;

  /* Byte 6 */
  tmp = (byte)((CID_Tab[1] & 0x0000FF00) >> 8);
  cardinfo->SD_cid.ProdName[3] = tmp;

  /* Byte 7 */
  tmp = (byte)(CID_Tab[1] & 0x000000FF);
  cardinfo->SD_cid.ProdName[4] = tmp;
  cardinfo->SD_cid.ProdName[5] = 0;

  /* Byte 8 */
  tmp = (byte)((CID_Tab[2] & 0xFF000000) >> 24);
  cardinfo->SD_cid.ProdRev[0] = '0'|(tmp>>4);
  cardinfo->SD_cid.ProdRev[1] = '.';
  cardinfo->SD_cid.ProdRev[2] = '0'|(tmp&15);
  cardinfo->SD_cid.ProdRev[3] = 0;

  /* Byte 9 */
  tmp = (byte)((CID_Tab[2] & 0x00FF0000) >> 16);
  cardinfo->SD_cid.ProdSN = tmp << 24;

  /* Byte 10 */
  tmp = (byte)((CID_Tab[2] & 0x0000FF00) >> 8);
  cardinfo->SD_cid.ProdSN |= tmp << 16;

  /* Byte 11 */
  tmp = (byte)(CID_Tab[2] & 0x000000FF);
  cardinfo->SD_cid.ProdSN |= tmp << 8;

  /* Byte 12 */
  tmp = (byte)((CID_Tab[3] & 0xFF000000) >> 24);
  cardinfo->SD_cid.ProdSN |= tmp;

  /* Byte 13 */
  tmp = (byte)((CID_Tab[3] & 0x00FF0000) >> 16);
  cardinfo->SD_cid.Reserved1 |= (tmp & 0xF0) >> 4;
  cardinfo->SD_cid.ManufactDate = (tmp & 0x0F) << 8;

  /* Byte 14 */
  tmp = (byte)((CID_Tab[3] & 0x0000FF00) >> 8);
  cardinfo->SD_cid.ManufactDate |= tmp;

  /* Byte 15 */
  tmp = (byte)(CID_Tab[3] & 0x000000FF);
  cardinfo->SD_cid.CID_CRC = (tmp & 0xFE) >> 1;
  cardinfo->SD_cid.Reserved2 = 1;
}



/* Enables wide bus opeartion for the requested card if supported by card */
/*    WideMode: Specifies the SD card wide bus mode.
  *   This parameter can be one of the following values:
  *     @arg SDIO_BusWide_4b: 4-bit data transfer
  *     @arg SDIO_BusWide_1b: 1-bit data transfer
  */
bool SD_EnableWideBusOperation (byte busbreite)
{ if (iCardType==MMC) return 0;   /* MMC Card doesn't support this feature */

  if ((iCardType==SD1_1) ||
      (iCardType==SD2_0) ||
      (iCardType==SDHC))
  { if (busbreite==4)              /* 4 Bit */
    { if (SwitchToWideBus())
      { SDIO_CLKCR = (1<<14) |       /* HW-Flusteuerung ein, um ber/Unterlufe zu vermeiden */
                     (1<<11) |       /* Busbreite = 4 */
                     (1<<8)  |       /* Clock Enable  */
                      2;             /* Clockrate ist 72MHz/(2+zahl), also hier 72/4=18MHz */

        return 1;
      }
    }
    else                           /* 1 Bit Modus */
    { if (SwitchTo1BitBus())
      { Setze_SDIO_CLKCR(8, 1);    /* war 0, SDIO_BusWide_1b */
        return 1;
      }
    }
  }
  return 0;
}




/************************************************************************************/
/*  Betriebsmode = Polling                                                          */
/************************************************************************************/

int GetBlockSizeID (word BlockSize)
{ int idx;
  if (!BlockSize) return -1;
  idx = 14;
  while (BlockSize <= 16384)
  { if (BlockSize == 16384) return idx;
    BlockSize = BlockSize << 1;
    --idx;
  }
  return -1;
}


/**
  * Allows to read one block from a specified address in a card.
  *   SecNum: Sector number
  *   readbuff: pointer to the buffer that will contain the received data
  *   BlockSize: the SD card Data block size.
  */
bool SD_ReadBlock(dword SecNum, dword *readbuff, dword BlockSize)
{ dword erg;
  dword addr;
  dword count;
  dword *tempbuff;
  int blidx;
  int i;

  if (!readbuff) return 0;

  tempbuff = readbuff;
  if (SDIO_RESP1 & SD_CARD_LOCKED) return 0;

  if (iCardType != SDHC)        /* Test auf eventuellen Adressberlauf */
  { if (SecNum & 0xFF800000) return 0;
    addr = SecNum * 512;
  }
  else
  { BlockSize = 512; addr = SecNum; }

  blidx = GetBlockSizeID(BlockSize);
  if (blidx < 0) return 0;

                             /* Set Block Size for Card */
  SDIO_SendCommand(16, WAITRESP_SHORT|CPSMEN, BlockSize);
  if (!CmdResp1Result(16)) return 0;

  SDIO_DTIMER = 0xFFFF;
  SDIO_DLEN   = BlockSize;
  SDIO_DCTRL  = 3 | (blidx<<4);   /* blen , rd, start */


                             /* Send CMD17 READ_SINGLE_BLOCK */
  SDIO_SendCommand(17, WAITRESP_SHORT|CPSMEN, addr);
  if (!CmdResp1Result(17)) return 0;

  count = 0;                                  /* Polling mode */
  while (!(SDIO_STA & (RXOVERR | DCRCFAIL | DTIMEOUT | DBCKEND | STBITERR)))
    { if (SDIO_STA & RXFIFOHF)
      { i = 8; while (i--) *tempbuff++ = SDIO_FIFO;
        count += 32;
      }
    }

  erg = SDIO_STA & (DTIMEOUT | DCRCFAIL | RXOVERR | STBITERR);
  if (erg)
    { SDIO_ICR = erg;
      if (erg & DTIMEOUT) String_Out("Timeout ",2);
      if (erg & DCRCFAIL) String_Out("CrcFail ",2);
      if (erg & RXOVERR)  String_Out("RXover ",2);
      if (erg & STBITERR) String_Out("StBitEr",2);

      String_Out(" err ",2); HexL_Out(erg,2); CRLF_Out(2);
      return 0;
    }

  while (SDIO_STA & RXDAVL)
    { *tempbuff = SDIO_FIFO;
      tempbuff++;
      count++;
    }

    /* Clear all the static flags */
  SDIO_ICR = ALL_STATIC_FLAGS;

  return 1;
}


/**
  * Allows to write one block starting from a specified address in a card.
  *   addr: Address from where data are to be read.
  *   writebuff: pointer to the buffer that contain the data to be transferred.
  *   BlockSize: the SD card Data block size.
  */
bool SD_WriteBlock(dword SecNum, dword *writebuff, dword BlockSize)
{ dword erg;
  dword addr;
  long  i;
  long  written;
  byte  cardstate;
  dword *tempbuff;
  int blidx;

  if (!writebuff) return 0;

  tempbuff = writebuff;
  SDIO_DCTRL &= 0xFFFFFF00;
  if (SDIO_RESP1 & SD_CARD_LOCKED) return 0;

  if (iCardType != SDHC)        /* Test auf eventuellen Adressberlauf */
  { if (SecNum & 0xFF800000) return 0;
    addr = SecNum * 512;
  }
  else
  { BlockSize = 512; addr = SecNum; }

  blidx = GetBlockSizeID(BlockSize);
  if (blidx < 0) return 0;

                             /* Set Block Size for Card */
  SDIO_SendCommand(16,WAITRESP_SHORT|CPSMEN,(dword) BlockSize);
  if (!CmdResp1Result(16)) return 0;

  erg = 0;                   /* Wait until card is ready for data Added */
  i = 64000;
  while ((i) && (!(erg & 0x100)))
  {  SDIO_SendCommand(13,WAITRESP_SHORT|CPSMEN,(dword) (RCA << 16));
     if (!CmdResp1Result(13)) return 0;
     erg = SDIO_RESP1;
  }
  if (!i) return 0;

  /* Send CMD24 WRITE_SINGLE_BLOCK */
  SDIO_SendCommand(24,WAITRESP_SHORT|CPSMEN,addr);
  if (!CmdResp1Result(24)) return 0;

/*  SDIO_DataConfig(0xFFFF, BlockSize, ((dword) blidx << 4) | DPSM_Enable); */
  SDIO_DTIMER = 0xFFFF;
  SDIO_DLEN   = BlockSize;
  SDIO_DCTRL  = 1 | (blidx<<4);   /* blen 8, wr, start */

                       /* In case of single data block transfer no need of stop command at all */
  written = 0;
  while (!(SDIO_STA & (DBCKEND | TXUNDERR | DCRCFAIL | DTIMEOUT | STBITERR)))
    { if (SDIO_STA &TXFIFOHE)
      { i = BlockSize - written;
        if (i < 32)
        { while (i > 0)
          { SDIO_FIFO = *tempbuff;
	    tempbuff++;
	    i -= 4;
            written += 4;
          }
        }
        else
        { for (i = 0; i < 8; i++)  { SDIO_FIFO = *(tempbuff + i);  }
          tempbuff += 8;
          written += 32;
        }
      }
    }

  erg = SDIO_STA & (DTIMEOUT | DCRCFAIL | TXUNDERR | STBITERR);
  if (erg) { SDIO_ICR = erg; return 0; }

  /* Clear all the static flags */
  SDIO_ICR = ALL_STATIC_FLAGS;

  /* warten, bis die Operation abgeschlossen ist */
  while (IsCardProgramming(&cardstate))
  { if ((cardstate != 6)&&(cardstate != 7)) return 1;  }
  return 0;
}


bool SD_DriveInit (void)
{ iCardType = 0;
  String_Out("disk_initialize\r\n",2);
   if (!SD_Init()) return 0;
   String_Out("translate\r\n",2);
   TranslateCardInfo(&SDCardInfo);    /* Read CSD/CID MSD registers */
                                      /* Select card */
   if (!SD_SelectDeselect((dword)(SDCardInfo.RCA << 16))) return 0;
   if (!SD_EnableWideBusOperation(4)) return 0;
   String_Out("wide bus ok\r\n",2);

  /* Set Device Transfer Mode to DMA */
  /*  if (!SD_SetDeviceMode(SD_DMA_MODE)) return STA_NODISK; */

  return 1;
}



